GOOGLE_URL = "google"
GOOGLE_ADS = "google_ads"
GOOGLE_HOTELS = "google_hotels"
GOOGLE_SEARCH = "google_search"
GOOGLE_IMAGES = "google_images"
GOOGLE_SUGGESTIONS = "google_suggest"
GOOGLE_TRAVEL_HOTELS = "google_travel_hotels"
GOOGLE_TRENDS_EXPLORE = "google_trends_explore"

BING_URL = "bing"
BING_SEARCH = "bing_search"

YANDEX_URL = "yandex"
YANDEX_SEARCH = "yandex_search"

BAIDU_URL = "baidu"
BAIDU_SEARCH = "baidu_search"

GOOGLE_SHOPPING_URL = "google_shopping"
GOOGLE_SHOPPING_SEARCH = "google_shopping_search"
GOOGLE_SHOPPING_PRODUCT = "google_shopping_product"
GOOGLE_SHOPPING_PRICING = "google_shopping_pricing"

WAYFAIR = "wayfair"
WAYFAIR_SEARCH = "wayfair_search"

UNIVERSAL = "universal_ecommerce"

AMAZON_URL = "amazon"
AMAZON_SEARCH = "amazon_search"
AMAZON_PRODUCT = "amazon_product"
AMAZON_PRICING = "amazon_pricing"
AMAZON_REVIEWS = "amazon_reviews"
AMAZON_QUESTIONS = "amazon_questions"
AMAZON_BEST_SELLERS = "amazon_bestsellers"
AMAZON_SELLERS = "amazon_sellers"



SYNC_BASE_URL = "https://realtime.oxylabs.io/v1/queries"
ASYNC_BASE_URL = "https://data.oxylabs.io/v1/queries"

PROXY_BASE_URL = "realtime.oxylabs.io"
PROXY_PORT = 60000
NON_UNIVERSAL_DOMAINS = {"google", "bing", "amazon", "wayfair"}


DEFAULT_REQUEST_TIMEOUT = 165
DEFAULT_POLL_INTERVAL = 5
DEFAULT_REQUEST_TIMEOUT_ASYNC = 105
DEFAULT_JOB_COMPLETION_TIMEOUT = 50



from typing import Optional

from oxylabs.sources.ecommerce.response import EcommerceResponse
from oxylabs.utils.types import source
from oxylabs.utils.utils import (
    check_parsing_instructions_validity,
    prepare_config,
)


class GoogleShopping:
    def __init__(self, ecommerce_instance) -> None:
        """
        Initializes an instance of the GoogleShopping class.

        Args:
            ecommerce_instance: The Ecommerce instance associated with the
            GoogleShopping class.
        """
        self._ecommerce_instance = ecommerce_instance

    def scrape_shopping_search(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        results_language: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        render: Optional[str] = None,
        parse: Optional[bool] = None,
        context: Optional[list] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping search results for a given query.

        Args:
            query (str): UTF-encoded keyword
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            results_language (Optional[str]): None,
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            render (Optional[str]): Enables JavaScript rendering.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SHOPPING_SEARCH,
            "domain": domain,
            "query": query,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "results_language": results_language,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "context": context,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._ecommerce_instance._get_resp(payload, config)
        return response

    def scrape_shopping_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        geo_location: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping search results for a given URL.

        Args:
            url (str): Direct URL (link) to Google page
            the search.
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            geo_location (Optional[str]): None,
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SHOPPING_URL,
            "url": url,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "geo_location": geo_location,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._ecommerce_instance._get_resp(payload, config)
        return response

    def scrape_shopping_products(
        self,
        query: str,
        domain: Optional[str] = None,
        locale: Optional[str] = None,
        results_language: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping product results for a given query.

        Args:
            query (str): UTF-encoded product code.
            domain (Optional[str]): The domain to limit the search results to.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            "results_language": None,
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SHOPPING_PRODUCT,
            "query": query,
            "domain": domain,
            "locale": locale,
            "results_language": results_language,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._ecommerce_instance._get_resp(payload, config)
        return response

    def scrape_product_pricing(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        results_language: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping product pricing results for a given product code.

        Args:
            query (str): UTF-encoded product code.
            domain (Optional[str]): The domain to limit the search results to.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            start_page (Optional[int]): The starting page number.
                        pages (Optional[int]): The number of pages to scrape.
            "results_language": None,
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SHOPPING_PRICING,
            "domain": domain,
            "query": query,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "results_language": results_language,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._ecommerce_instance._get_resp(payload, config)
        return response


class GoogleShoppingAsync:
    def __init__(self, ecommerce_async_instance) -> None:
        """
        Initializes an instance of the GoogleShoppingAsync class.

        Args:
            ecommerce_async_instance: The EcommerceAsync instance associated
            with the GoogleShoppingAsync class.
        """
        self._ecommerce_async_instance = ecommerce_async_instance

    async def scrape_shopping_search(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        results_language: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        render: Optional[str] = None,
        parse: Optional[bool] = None,
        context: Optional[list] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping search results for a given query.

        Args:
            query (str): UTF-encoded keyword.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
                        pages (Optional[int]): The number of pages to scrape.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            "results_language": None,
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            render (Optional[str]): Enables JavaScript rendering.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int, optional): The interval in seconds for the
            request to poll the server for a response. Defaults to 5.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SHOPPING_SEARCH,
            "domain": domain,
            "query": query,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "results_language": results_language,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "context": context,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response

    async def scrape_shopping_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        geo_location: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping search results for a given URL.

        Args:
            url (str): Direct URL (link) to Google page.
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            geo_location (Optional[str]): None,
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds
            for the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int, optional): The interval in seconds for the
            request to poll the server for a response. Defaults to 5.
        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SHOPPING_URL,
            "url": url,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "geo_location": geo_location,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response

    async def scrape_shopping_products(
        self,
        query: str,
        domain: Optional[str] = None,
        locale: Optional[str] = None,
        results_language: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping product results for a given query.

        Args:
            query (str): UTF-encoded product code.
            domain (Optional[str]): The domain to limit the search results to.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            "results_language": None,
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int, optional): The interval in seconds for the
            request to poll the server for a response. Defaults to 5.
        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SHOPPING_PRODUCT,
            "query": query,
            "domain": domain,
            "locale": locale,
            "results_language": results_language,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response

    async def scrape_product_pricing(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        results_language: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Google Shopping product pricing results for a given product code.

        Args:
            url (str): UTF-encoded product code.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
                        pages (Optional[int]): The number of pages to scrape.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            "results_language": None,
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int, optional): The interval in seconds for the
            request to poll the server for a response. Defaults to 5.
        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SHOPPING_PRICING,
            "domain": domain,
            "query": query,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "results_language": results_language,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response



class SERPResponse:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.results = [Results(item) for item in data.get("results", [])]
        self.job = Job(data.get("job", {}))


class Results:
    def __init__(self, data):
        if data is None:
            data = {}
        self.custom_content_parsed = data.get("custom_content_parsed", {})
        self.content_parsed = Content(data.get("content_parsed", {}))
        self.content = data.get("content")
        self.created_at = data.get("created_at")
        self.updated_at = data.get("updated_at")
        self.page = data.get("page")
        self.url = data.get("url")
        self.job_id = data.get("job_id")
        self.status_code = data.get("status_code")
        self.parser_type = data.get("parser_type")


class Content:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.page = data.get("page")
        self.errors = data.get("_errors")
        self.results = Result(data.get("results", {}))
        self.last_visible_page = data.get("last_visible_page")
        self.parse_status_code = data.get("parse_status_code")


class Result:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pla = Pla(data.get("pla", {}))
        self.paid = [Paid(item) for item in data.get("paid", [])]
        self.images = Image(data.get("images", {}))
        self.organic = [Organic(item) for item in data.get("organic", [])]
        self.twitter = Twitter(data.get("twitter", {}))
        self.knowledge = Knowledge(data.get("knowledge", {}))
        self.local_pack = LocalPack(data.get("local_pack", {}))
        self.top_stories = TopStory(data.get("top_stories", {}))
        self.popular_products = [
            PopularProducts(item) for item in data.get("popular_products", [])
        ]
        self.related_searches = RelatedSearches(
            data.get("related_searches", {})
        )
        self.related_questions = RelatedQuestions(
            data.get("related_questions", {})
        )
        self.search_information = SearchInformation(
            data.get("search_information", {})
        )
        self.item_carousel = ItemCarousel(data.get("item_carousel", {}))
        self.recipes = Recipes(data.get("recipes", {}))
        self.videos = Videos(data.get("videos", {}))
        self.featured_snippet = [
            FeaturedSnippet(item) for item in data.get("featured_snippet", [])
        ]
        self.related_searches_categorized = [
            RelatedSearchesCategorized(item)
            for item in data.get("related_searches_categorized", [])
        ]
        self.hotels = Hotels(data.get("hotels", {}))
        self.flights = Flights(data.get("flights", {}))
        self.video_box = VideoBox(data.get("video_box", {}))
        self.local_service_ads = LocalServiceAds(
            data.get("local_service_ads", {})
        )
        self.navigation = [
            Navigation(item) for item in data.get("navigation", [])
        ]
        self.instant_answers = [
            InstantAnswers(item) for item in data.get("instant_answers", [])
        ]
        self.visually_similar_images = VisuallySimilarImages(
            data.get("visually_similar_images", {})
        )
        self.total_results_count = data.get("total_results_count")


class Pla:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [PlaItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")


class PlaItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.price = data.get("price")
        self.title = data.get("title")
        self.seller = data.get("seller")
        self.url_image = data.get("url_image")
        self.image_data = data.get("image_data")


class Paid:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")
        self.data_rw = data.get("data_rw")
        self.data_pcu = data.get("data_pcu", [])
        self.sitelinks = PaidSitelinks(data.get("sitelinks", {}))
        self.url_shown = data.get("url_shown")
        self.pos_overall = data.get("pos_overall")


class PaidSitelinks:
    def __init__(self, data):
        if data is None:
            data = {}
        self.expanded = [Expanded(item) for item in data.get("expanded", [])]
        self.inline = [Inline(item) for item in data.get("inline", [])]


class Expanded:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")


class Inline:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")


class Image:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [ImageItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")


class ImageItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.alt = data.get("alt")
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.data = data.get("data")
        self.source = data.get("source")


class Organic:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")
        self.images = [item for item in data.get("images", [])]
        self.site_links = OrganicSitelinks(data.get("sitelinks", {}))
        self.url_shown = data.get("url_shown")
        self.pos_overall = data.get("pos_overall")


class OrganicSitelinks:
    def __init__(self, data):
        if data is None:
            data = {}
        self.expanded = [Expanded(item) for item in data.get("expanded", [])]
        self.inline = [Inline(item) for item in data.get("inline", [])]


class Twitter:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.items = [TwitterItem(item) for item in data.get("items", [])]
        self.title = data.get("title")
        self.pos_overall = data.get("pos_overall")


class TwitterItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.content = data.get("content")
        self.time_frame = data.get("time_frame")


class Knowledge:
    def __init__(self, data):
        if data is None:
            data = {}
        self.title = data.get("title")
        self.images = [item for item in data.get("images", [])]
        self.factoids = [Factoid(item) for item in data.get("factoids", [])]
        self.profiles = [Profile(item) for item in data.get("profiles", [])]
        self.subtitle = data.get("subtitle")
        self.description = data.get("description")
        self.related_searches = [
            RelatedSearches(item) for item in data.get("related_searches", [])
        ]


class Factoid:
    def __init__(self, data):
        if data is None:
            data = {}
        self.links = [LinkElement(item) for item in data.get("links", [])]
        self.title = data.get("title")
        self.content = data.get("content")


class LinkElement:
    def __init__(self, data):
        if data is None:
            data = {}
        self.href = data.get("href")
        self.title = data.get("title")


class Profile:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.title = data.get("title")


class RelatedSearches:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.title = data.get("title")
        self.section_title = data.get("section_title")


class LocalPack:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [LocalPackItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")


class LocalPackItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.cid = data.get("cid")
        self.pos = data.get("pos")
        self.links = [LocalPackLink(item) for item in data.get("links", [])]
        self.phone = data.get("phone")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.address = data.get("address")
        self.subtitle = data.get("subtitle")
        self.rating_count = data.get("rating_count")


class LocalPackLink:
    def __init__(self, data):
        if data is None:
            data = {}
        self.href = data.get("href")
        self.title = data.get("title")


class TopStory:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [TopStoryItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")


class TopStoryItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.title = data.get("title")
        self.source = data.get("source")
        self.time_frame = data.get("time_frame")


class PopularProducts:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.price = data.get("price")
        self.rating = data.get("rating")
        self.seller = data.get("seller")
        self.title = data.get("title")
        self.image_data = data.get("image_data")


class RelatedSearches:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos_overall = data.get("pos_overall")
        self.related_searches = [
            item for item in data.get("related_searches", [])
        ]


class RelatedQuestions:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos_overall = data.get("pos_overall")
        self.related_questions = [
            RelatedQuestionsItem(item)
            for item in data.get("related_questions", [])
        ]


class RelatedQuestionsItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.answer = data.get("answer")
        self.source = data.get("source")
        self.question = data.get("question")


class Source:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.title = data.get("title")
        self.url_shown = data.get("url_shown")


class SearchInformation:
    def __init__(self, data):
        if data is None:
            data = {}
        self.image = SearchInformationImage(data.get("image", {}))
        self.query = data.get("query")
        self.showing_results_for = data.get("showing_results_for")
        self.total_results_count = data.get("total_results_count")


class SearchInformationImage:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.width = data.get("width")
        self.height = data.get("height")
        self.other_sizes = data.get("other_sizes")


class ItemCarousel:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [ItemCarouselItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")
        self.title = data.get("title")


class ItemCarouselItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.href = data.get("href")
        self.title = data.get("title")
        self.subtitle = data.get("subtitle")


class Recipes:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [RecipesItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")


class RecipesItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.source = data.get("source")
        self.duration = data.get("duration")


class Videos:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [VideosItem(item) for item in data.get("items", [])]
        self.pos_overall = data.get("pos_overall")


class VideosItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.title = data.get("title")
        self.author = data.get("author")
        self.source = data.get("source")


class FeaturedSnippet:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")
        self.url_shown = data.get("url_shown")
        self.pos_overall = data.get("pos_overall")


class RelatedSearchesCategorized:
    def __init__(self, data):
        if data is None:
            data = {}
        self.items = [
            RelatedSearchesCategorizedItem(item)
            for item in data.get("items", [])
        ]
        self.category = data.get("category")
        self.pos_overall = data.get("pos_overall")


class RelatedSearchesCategorizedItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.title = data.get("title")


class Category:
    def __init__(self, data):
        if data is None:
            data = {}
        self.name = data.get("name")
        self.type = data.get("type")


class Hotels:
    def __init__(self, data):
        if data is None:
            data = {}
        self.date_to = data.get("date_to")
        self.results = [HotelsResult(item) for item in data.get("results", [])]
        self.date_from = data.get("date_from")
        self.pos_overall = data.get("pos_overall")


class HotelsResult:
    def __init__(self, data):
        if data is None:
            data = {}
        self.price = data.get("price")
        self.title = data.get("title")
        self.from_location = data.get("from")


class Flights:
    def __init__(self, data):
        if data is None:
            data = {}
        self.to = data.get("to")
        self.from_location = data.get("from")
        self.results = [
            FlightsResult(item) for item in data.get("results", [])
        ]
        self.date_from = data.get("date_from")
        self.pos_overall = data.get("pos_overall")


class FlightsResult:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.type = data.get("type")
        self.price = data.get("price")
        self.airline = data.get("airline")
        self.duration = data.get("duration")


class VideoBox:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.title = data.get("title")
        self.pos_overall = data.get("pos_overall")


class LocalServiceAds:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos_overall = data.get("pos_overall")
        self.items = [
            LocalServiceAdsItem(item) for item in data.get("items", [])
        ]


class LocalServiceAdsItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.reviews_count = data.get("reviews_count")
        self.google_gauranteed = data.get("google_gauranteed")


class Navigation:
    def __init__(self, data):
        if data is None:
            data = {}
        self.url = data.get("url")
        self.title = data.get("title")
        self.pos = data.get("pos")


class InstantAnswers:
    def __init__(self, data):
        if data is None:
            data = {}
        self.type = data.get("type")
        self.parsed = data.get("_parsed")
        self.pos_overall = data.get("pos_overall")


class VisuallySimilarImages:
    def __init__(self, data):
        if data is None:
            data = {}
        self.all_images_url = data.get("all_images_url")
        self.featured_images = data.get("featured_images")


class Job:
    def __init__(self, data):
        if data is None:
            data = {}
        self.callback_url = data.get("callback_url")
        self.client_id = data.get("client_id")
        self.context = [Context(item) for item in data.get("context", [])]
        self.created_at = data.get("created_at")
        self.domain = data.get("domain")
        self.geo_location = data.get("geo_location")
        self.id = data.get("id")
        self.limit = data.get("limit")
        self.locale = data.get("locale")
        self.pages = data.get("pages")
        self.parse = data.get("parse")
        self.parser_type = data.get("parser_type")
        self.parsing_instructions = data.get("parsing_instructions")
        self.browser_instructions = data.get("browser_instructions")
        self.render = data.get("render")
        self.url = data.get("url")
        self.query = data.get("query")
        self.source = data.get("source")
        self.start_page = data.get("start_page")
        self.status = data.get("status")
        self.storage_type = data.get("storage_type")
        self.storage_url = data.get("storage_url")
        self.subdomain = data.get("subdomain")
        self.content_encoding = data.get("content_encoding")
        self.updated_at = data.get("updated_at")
        self.user_agent_type = data.get("user_agent_type")
        self.session_info = data.get("session_info")
        self.statuses = data.get("statuses")
        self.client_notes = data.get("client_notes")
        self.links = [JobLink(item) for item in data.get("links", [])]


class Context:
    def __init__(self, data):
        if data is None:
            data = {}
        self.key = data.get("key")
        self.value = data.get("value")


class JobLink:
    def __init__(self, data):
        if data is None:
            data = {}
        self.rel = data.get("rel")
        self.href = data.get("href")
        self.method = data.get("method")



from typing import Optional

from oxylabs.sources.ecommerce.response import EcommerceResponse
from oxylabs.utils.types import source
from oxylabs.utils.utils import prepare_config


class Wayfair:
    def __init__(self, ecommerce_instance) -> None:
        """
        Initializes an instance of the Wayfair class.

        Args:
            ecommerce_instance: The Ecommerce instance associated with the
            Wayfair class.
        """
        self._ecommerce_instance = ecommerce_instance

    def scrape_search(
        self,
        query: str,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        limit: Optional[int] = None,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        request_timeout: Optional[int] = 165,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Wayfair search results for a given query.

        Args:
            query (str): The search query.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            limit (Optional[int]): Number of results to retrieve in each page.
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.WAYFAIR_SEARCH,
            "query": query,
            "start_page": start_page,
            "pages": pages,
            "limit": limit,
            "user_agent_type": user_agent_type,
            "callback_url": callback_url,
            **kwargs,
        }
        response = self._ecommerce_instance._get_resp(payload, config)
        return response

    def scrape_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        request_timeout: Optional[int] = 165,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Wayfair search results for a given URL.

        Args:
            url (str): The URL to be scraped.
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.WAYFAIR,
            "url": url,
            "user_agent_type": user_agent_type,
            "callback_url": callback_url,
            **kwargs,
        }
        response = self._ecommerce_instance._get_resp(payload, config)
        return response


class WayfairAsync:
    def __init__(self, ecommerce_async_instance) -> None:
        """
        Initializes an instance of the WayfairAsync class.

        Args:
            ecommerce_async_instance: The EcommerceAsync instance associated
            with the WayfairAsync class.
        """
        self._ecommerce_async_instance = ecommerce_async_instance

    async def scrape_search(
        self,
        query: str,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        limit: Optional[int] = None,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Asynchronously scrapes Wayfair search results for a given query.

        Args:
            query (str): The search query.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            limit (Optional[int]): Number of results to retrieve in each page
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.WAYFAIR_SEARCH,
            "query": query,
            "start_page": start_page,
            "pages": pages,
            "limit": limit,
            "user_agent_type": user_agent_type,
            "callback_url": callback_url,
            **kwargs,
        }
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response

    async def scrape_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Asynchronously scrapes Wayfair search results for a given URL.

        Args:
            url (str): The URL to be scraped.
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.WAYFAIR,
            "url": url,
            "user_agent_type": user_agent_type,
            "callback_url": callback_url,
            **kwargs,
        }
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response



import asyncio
import base64
import logging

import aiohttp
import requests

from oxylabs.sources.ecommerce.ecommerce import Ecommerce, EcommerceAsync
from oxylabs.sources.serp.serp import SERP, SERPAsync
from oxylabs.utils.defaults import ASYNC_BASE_URL, SYNC_BASE_URL

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class APICredentials:
    def __init__(self, username: str, password: str) -> None:
        """
        Initializes an instance of ApiCredentials.

        Args:
            username (str): The username for API authentication.
            password (str): The password for API authentication.
        """
        self.username = username
        self.password = password

    def get_encoded_credentials(self) -> str:
        """
        Returns the Base64 encoded username and password for API
        authentication.
        """
        credentials = f"{self.username}:{self.password}"
        return base64.b64encode(credentials.encode()).decode()


class BaseClient:
    def __init__(self, base_url: str, api_credentials: APICredentials) -> None:
        self._base_url = base_url
        self._api_credentials = api_credentials
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self._api_credentials.get_encoded_credentials()}",
        }


class RealtimeClient(BaseClient):
    def __init__(self, username: str, password: str) -> None:
        super().__init__(SYNC_BASE_URL, APICredentials(username, password))
        self.serp = SERP(self)
        self.ecommerce = Ecommerce(self)

    def _req(self, payload: dict, method: str, config: dict) -> dict:
        """
        Sends a HTTP request to the specified URL with the given payload
        and method.

        Args:
            payload (dict): The payload to be sent with the request.
            method (str): The HTTP method to be used for the request
            (e.g., "POST", "GET").
            config (dict): Additional configuration options for the
            request.

        Returns:
            dict: The JSON response from the server, if the request is
            successful.
                  None, if an error occurs during the request.

        Raises:
            requests.exceptions.Timeout: If the request times out.
            requests.exceptions.HTTPError: If an HTTP error occurs.
            requests.exceptions.RequestException: If a general request
            error occurs.
        """
        try:
            if method == "POST":
                response = requests.post(
                    self._base_url,
                    headers=self._headers,
                    json=payload,
                    timeout=config["request_timeout"],
                )
            else:
                logger.error(f"Unsupported method: {method}")
                return None
            response.raise_for_status()

            if response.status_code == 200:
                return response.json()
            else:
                logger.error(f"Error occurred: {response.status_code}")
                return None

        except requests.exceptions.Timeout:
            logger.error(
                f"Timeout error. The request to {self._base_url} with method {method} has timed out."
            )
            return None
        except requests.exceptions.HTTPError as err:
            logger.error(f"HTTP error occurred: {err}")
            logger.error(response.text)
            return None
        except requests.exceptions.RequestException as err:
            logger.error(f"Error occurred: {err}")
            return None


class AsyncClient(BaseClient):
    def __init__(self, username: str, password: str) -> None:
        super().__init__(ASYNC_BASE_URL, APICredentials(username, password))
        self.serp = SERPAsync(self)
        self.ecommerce = EcommerceAsync(self)

    async def _get_job_id(
        self,
        payload: dict,
        user_session: aiohttp.ClientSession,
        request_timeout: int,
    ) -> str:
        try:
            async with user_session.post(
                self._base_url,
                headers=self._headers,
                json=payload,
                timeout=request_timeout,
            ) as response:
                data = await response.json()
                response.raise_for_status()
                return data["id"]
        except aiohttp.ClientResponseError as e:
            logger.error(
                f"HTTP error occurred: {e.status} - {e.message} - {data['message']}"
            )
        except aiohttp.ClientConnectionError as e:
            logger.error(f"Connection error occurred: {e}")
        except asyncio.TimeoutError:
            logger.error(
                f"Timeout error. The request to {self._base_url} has timed out."
            )
        except Exception as e:
            logger.error(f"Error occurred: {str(e)}")
            return None

    async def _poll_job_status(
        self,
        job_id: str,
        poll_interval: int,
        user_session: aiohttp.ClientSession,
        timeout: int,
    ) -> bool:
        job_status_url = f"{self._base_url}/{job_id}"
        end_time = asyncio.get_event_loop().time() + timeout
        while asyncio.get_event_loop().time() < end_time:
            try:
                async with user_session.get(
                    job_status_url,
                    headers=self._headers,
                    timeout=poll_interval,
                ) as response:
                    data = await response.json()
                    response.raise_for_status()
                    if data["status"] == "done":
                        return True
                    elif data["status"] == "faulted":
                        raise Exception("Job faulted")
            except Exception as e:
                logger.error(f"Error occurred: {str(e)}")
                return False
            await asyncio.sleep(poll_interval)

        logger.info("Job completion timeout exceeded")
        return False

    async def _get_http_resp(
        self, job_id: str, user_session: aiohttp.ClientSession
    ) -> dict:
        """
        Retrieves the HTTP response for a given job ID.

        Args:
            job_id (str): The ID of the job.
            user_session (aiohttp.ClientSession): The client session used for
            making the request.

        Returns:
            dict: The JSON response data.

        Raises:
            aiohttp.ClientResponseError: If a client response error occurs.
            aiohttp.ClientConnectionError: If a client connection error occurs.
            asyncio.TimeoutError: If the request times out.
            Exception: If any other error occurs.
        """
        result_url = f"{self._base_url}/{job_id}/results"
        try:
            async with user_session.get(
                result_url, headers=self._headers
            ) as response:
                data = await response.json()
                response.raise_for_status()
                return data
        except aiohttp.ClientResponseError as e:
            logger.error(
                f"HTTP error occurred: {e.status} - {e.message} - {data['message']}"
            )
        except aiohttp.ClientConnectionError as e:
            logger.error(f"Connection error occurred: {e}")
        except asyncio.TimeoutError:
            logger.error(
                f"Timeout error. The request to {result_url} has timed out."
            )
        except Exception as e:
            logger.error(f"An error occurred: {e} - {data['message']}")
        return None

    async def _execute_with_timeout(
        self, payload: dict, config: dict, user_session: aiohttp.ClientSession
    ) -> dict:

        request_timeout = config["request_timeout"]
        job_completion_timeout = config["job_completion_timeout"]
        poll_interval = config["poll_interval"]

        job_id = await self._get_job_id(payload, user_session, request_timeout)
        if not job_id:
            logger.error("Failed to get job ID")

        job_completed = await self._poll_job_status(
            job_id, poll_interval, user_session, job_completion_timeout
        )
        if not job_completed:
            logger.error("Job did not complete successfully")

        result = await self._get_http_resp(job_id, user_session)
        return result



from typing import Optional

from oxylabs.sources.serp.response import SERPResponse
from oxylabs.utils.types import source
from oxylabs.utils.utils import (
    check_parsing_instructions_validity,
    prepare_config,
)


class Google:
    def __init__(self, serp_instance) -> None:
        """
        Initializes an instance of the Google class.

        Args:
            serp_instance: The SERP instance associated with the Google class.
        """
        self._serp_instance = serp_instance

    def scrape_search(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        limit: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
                        pages (Optional[int]): The number of pages to scrape.
            limit (Optional[int]): Number of results to retrieve in each page.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SEARCH,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "limit": limit,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "context": context,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        geo_location: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google search results for a given URL.

        Args:
            url (str): The URL to be scraped.
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            render (Optional[str]): Enables JavaScript rendering.
            geo_location (Optional[str]): None,
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_URL,
            "url": url,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "geo_location": geo_location,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)

        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_ads(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google Ads search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_ADS,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            "context": context,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_suggestions(
        self,
        query: str,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google suggestions for a given query.

        Args:
            query (str): The search query.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SUGGESTIONS,
            "query": query,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            **kwargs,
        }
        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_hotels(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        limit: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google Hotels search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            limit (Optional[int]): Number of results to retrieve in each page.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_HOTELS,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "limit": limit,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "context": context,
            **kwargs,
        }
        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_travel_hotels(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google Travel Hotels search results for a given query.

        Args:
            query (str): UTF-encoded keyword. "query": "hotels" will result
            list hotels in a given geo_location; "query": "hotels in
            <Location>" will result in a list hotels for <Location>. Eg hotels
            in Paris will list hotels in Paris, no matter what geo_location is
            given.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_TRAVEL_HOTELS,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "context": context,
            **kwargs,
        }
        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_images(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google Images search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_SEARCH,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            "context": context,
            **kwargs,
        }
        payload.setdefault("context", [])

        for item in payload["context"]:
            if item.get("key") == "tbm":
                item["value"] = item.get("value", "isch")
                break
        else:
            payload["context"].append({"key": "tbm", "value": "isch"})

        check_parsing_instructions_validity(parsing_instructions)
        response = self._serp_instance._get_resp(payload, config)
        return response

    def scrape_trends_explore(
        self,
        query: str,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        **kwargs,
    ) -> SERPResponse:
        """
        Scrapes Google Trends Explore results for a given query.

        Args:
            query (str): The search query.
            user_agent_type (Optional[str]): Device type and browser.
            geo_location (Optional[str]): None,
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.GOOGLE_TRENDS_EXPLORE,
            "query": query,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "callback_url": callback_url,
            "context": context,
            **kwargs,
        }
        response = self._serp_instance._get_resp(payload, config)
        return response


class GoogleAsync:
    def __init__(self, serp_async_instance) -> None:
        """
        Initializes an asynchronous Google object.

        Args:
            serp_async_instance: The SERP instance to be used for Google
            requests.
        """

        self._serp_async_instance = serp_async_instance

    async def scrape_search(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        limit: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google search results for a given query.

        Args:

            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            limit (Optional[int]): Number of results to retrieve in each page.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SEARCH,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "limit": limit,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "context": context,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_url(
        self,
        url: str,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        geo_location: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google search results for a given URL.

        Args:
            url (str): The URL to be scraped.
            user_agent_type (Optional[str]): Device type and browser.
            callback_url (Optional[str]): URL to your callback endpoint.
            render (Optional[str]): Enables JavaScript rendering.
            geo_location (Optional[str]): None,
            parse (Optional[bool]): true will return structured data.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_URL,
            "url": url,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "geo_location": geo_location,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)

        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_ads(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google Ads search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_ADS,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            "context": context,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_suggestions(
        self,
        query: str,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google suggestions for a given query.

        Args:
            query (str): The search query.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SUGGESTIONS,
            "query": query,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            **kwargs,
        }
        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_hotels(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        limit: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google Hotels search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            limit (Optional[int]): Number of results to retrieve in each page.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_HOTELS,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "limit": limit,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "context": context,
            **kwargs,
        }
        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_travel_hotels(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google Travel Hotels search results for a given
        query.

        Args:
            query (str): UTF-encoded keyword. "query": "hotels" will result
            list hotels in a given geo_location;
            "query": "hotels in <Location>" will result in a list hotels for
            <Location>. Eg hotels in Paris will list hotels in Paris, no matter
            what geo_location is given.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            user_agent_type (Optional[str]): Device type and browser.
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_TRAVEL_HOTELS,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "context": context,
            **kwargs,
        }
        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_images(
        self,
        query: str,
        domain: Optional[str] = None,
        start_page: Optional[int] = None,
        pages: Optional[int] = None,
        locale: Optional[str] = None,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        render: Optional[str] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parsing_instructions: Optional[dict] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google Images search results for a given query.

        Args:
            query (str): The search query.
            domain (Optional[str]): The domain to limit the search results to.
            start_page (Optional[int]): The starting page number.
            pages (Optional[int]): The number of pages to scrape.
            user_agent_type (Optional[str]): Device type and browser.
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            geo_location (Optional[str]): None,
            render (Optional[str]): Enables JavaScript rendering.
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            context: Optional[list],
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_SEARCH,
            "query": query,
            "domain": domain,
            "start_page": start_page,
            "pages": pages,
            "locale": locale,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "render": render,
            "callback_url": callback_url,
            "parse": parse,
            "parsing_instructions": parsing_instructions,
            "context": context,
            **kwargs,
        }
        payload.setdefault("context", [])

        for item in payload["context"]:
            if item.get("key") == "tbm":
                item["value"] = item.get("value", "isch")
                break
        else:
            payload["context"].append({"key": "tbm", "value": "isch"})

        check_parsing_instructions_validity(parsing_instructions)
        response = await self._serp_async_instance._get_resp(payload, config)
        return response

    async def scrape_trends_explore(
        self,
        query: str,
        geo_location: Optional[str] = None,
        user_agent_type: Optional[str] = None,
        callback_url: Optional[str] = None,
        context: Optional[list] = None,
        request_timeout: Optional[int] = 165,
        job_completion_timeout: Optional[int] = None,
        poll_interval: Optional[int] = None,
        **kwargs,
    ) -> SERPResponse:
        """
        Asynchronously scrapes Google Trends Explore results for a given query.

        Args:
            query (str): The search query.
            user_agent_type (Optional[str]): Device type and browser.
            geo_location (Optional[str]): None,
            callback_url (Optional[str]): URL to your callback endpoint.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50

        Returns:
            SERPResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.GOOGLE_TRENDS_EXPLORE,
            "query": query,
            "geo_location": geo_location,
            "user_agent_type": user_agent_type,
            "callback_url": callback_url,
            "context": context,
            **kwargs,
        }
        response = await self._serp_async_instance._get_resp(payload, config)
        return response



import logging
from typing import Optional
from urllib.parse import quote, urlparse

import requests

from oxylabs.utils.defaults import (
    NON_UNIVERSAL_DOMAINS,
    PROXY_BASE_URL,
    PROXY_PORT,
)
from oxylabs.utils.utils import prepare_config

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ProxyClient:
    def __init__(self, username: str, password: str) -> None:
        """
        Initializes a ProxyClient object with the provided username and password.

        Args:
            username (str): The username for the proxy authentication.
            password (str): The password for the proxy authentication.
        """
        self._username = quote(username)
        self._password = quote(password)
        self._proxy_url = self._build_proxy_url()
        self._session = requests.Session()
        self._session.proxies = {
            "http": self._proxy_url,
            "https": self._proxy_url,
        }
        self._session.verify = False
        self._url_to_scrape = None

    def _build_proxy_url(self) -> str:
        """
        Build the proxy URL using configured constants.

        Returns:
            str: The constructed proxy URL.
        """
        return f"http://{self._username}:{self._password}@{PROXY_BASE_URL}:{PROXY_PORT}"

    def get(
        self, url: str, request_timeout: Optional[int] = None
    ) -> Optional[requests.Response]:
        """
        Sends a GET request to the specified URL using the session object.

        Args:
            url (str): The URL to send the GET request to.
            request_timeout (Optional[int]): The request timeout in seconds.
            Defaults to None (no timeout).

        Returns:
            Optional[requests.Response]: The response object returned by the
            GET request, or None if an error occurred.
        """
        try:
            config = prepare_config(request_timeout=request_timeout)
            self._url_to_scrape = url
            response = self._session.get(
                url, timeout=config["request_timeout"]
            )
            response.raise_for_status()
            return response
        except requests.exceptions.Timeout:
            logger.error(
                f"Timeout error. The request to {url} has timed out after {request_timeout} seconds."
            )
            return None
        except requests.exceptions.RequestException as e:
            logger.error(f"Request failed: {e}")
            return None

    def add_user_agent_header(self, user_agent_type: str) -> None:
        """
        Adds a user agent header to the session headers.
        There is no way to indicate a specific User-Agent, but you can let us
        know which user-agent type you would like us to use.

        Args:
            user_agent_type (str): The user agent to add. Must be one of the
            following:
            - "desktop"         - A User-Agent of a desktop browser.
            - "desktop_chrome"  - A User-Agent of one of the latest versions of
              a desktop Chrome browser.
            - "desktop_edge"    - A User-Agent of one of the latest versions of
              a desktop Edge browser.
            - "desktop_firefox" - A User-Agent of one of the latest versions of
              a desktop Firefox browser.
            - "desktop_opera"   - A User-Agent of one of the latest versions of
              a desktop Opera browser.
            - "desktop_safari"  - A User-Agent of one of the latest versions of
              a desktop Safari browser.
            - "mobile"          - A User-Agent of a mobile browser.
            - "mobile_android"  - A User-Agent of one of the latest versions of
              an Android mobile browser.
            - "mobile_ios"      - A User-Agent of one of the latest versions of
              an iOS mobile browser.
            - "tablet"          - A User-Agent of a tablet browser.
            - "tablet_android"  - A User-Agent of one of the latest versions of
              an Android tablet browser.
            - "tablet_ios"      - A User-Agent of one of the latest versions of
              an iOS tablet browser.

        Returns:
            None
        """
        self._session.headers["x-oxylabs-user-agent-type"] = user_agent_type

    def add_render_header(self, render: str) -> None:
        """
        Adds a render header to the session headers.

        Args:
            render (str): The render type to add. Must be one of the following:
            - "html" - The output will include an HTML result.
            - "png"  - The output will include a PNG screenshot of the result.

        Returns:
            None
        """
        self._session.headers["x-oxylabs-render"] = render

    def add_parse_header(
        self, parse: bool = False, parsing_instructions: Optional[dict] = None
    ) -> None:
        """
        Adds a parse header to the session headers.

        Args:
            parse (bool, optional): Whether to enable parsing. Defaults to
            False.
            parsing_instructions (dict, optional): Instructions for parsing.
            Defaults to None.

        Returns:
            None
        """

        if parse or parsing_instructions:
            self._session.headers["x-oxylabs-parse"] = "1"
            if self._is_universal_source():
                self._session.headers["x-oxylabs-parser-type"] = (
                    "universal_ecommerce"
                )
            else:
                self._session.headers.pop("x-oxylabs-parser-type", None)
        else:
            self._session.headers.pop("x-oxylabs-parse", None)

    def _is_universal_source(self) -> bool:
        """
        Checks if the URL to scrape belongs to a universal source.

        Returns:
            bool: True if the URL belongs to a universal source, False
            otherwise.
        """
        parsed_url = urlparse(self._url_to_scrape)
        if any(
            domain in parsed_url.netloc.decode()
            for domain in NON_UNIVERSAL_DOMAINS
        ):
            return False

        return True

    def add_geo_location_header(self, geo_location: str) -> None:
        """
        Adds a geo location header to the session headers.
        In some cases, you may need to indicate the geographical location that
        the result should be adapted for.

        Args:
            geo_location (str): The geo location to add. Accepted values depend
            on the URL you would like us to scrape.

        Returns:
            None
        """
        self._session.headers["x-oxylabs-geo-location"] = geo_location



class EcommerceResponse:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.results = [Results(item) for item in data.get("results", [])]
        self.job = Job(data.get("job", {}))


class Results:
    def __init__(self, data):
        if data is None:
            data = {}
        self.custom_content_parsed = data.get("custom_content_parsed", {})
        self.content_parsed = Content(data.get("content_parsed", {}))
        self.content = data.get("content")
        self.created_at = data.get("created_at")
        self.updated_at = data.get("updated_at")
        self.page = data.get("page")
        self.url = data.get("url")
        self.job_id = data.get("job_id")
        self.status_code = data.get("status_code")
        self.parser_type = data.get("parser_type")


class Content:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.title = data.get("title")
        self.pages = data.get("pages")
        self.query = data.get("query")
        self.images = data.get("images")
        self.variants = Variants(data.get("variants", {}))
        self.highlights = data.get("highlights", [])
        self.description = data.get("description")
        self.related_items = RelatedItems(data.get("related_items", {}))
        self.specifications = Specifications(data.get("specifications", {}))
        self.page = data.get("page")
        self.errors = data.get("_errors")
        self.results = Result(data.get("results", {}))
        self.rating = data.get("rating")
        self.pricing = [Pricing(item) for item in data.get("pricing", [])]
        self.ads = [AmazonProductAds(item) for item in data.get("ads", [])]
        self.asin = data.get("asin")
        self.price = data.get("price")
        self.stock = data.get("stock")
        self.coupon = data.get("coupon")
        self.category = [
            AmazonProductCategory(item) for item in data.get("category", [])
        ]
        self.currency = data.get("currency")
        self.delivery = [
            AmazonProductDelivery(item) for item in data.get("delivery", [])
        ]
        self.warnings = data.get("_warnings", [])
        self.deal_type = data.get("deal_type")
        self.page_type = data.get("page_type")
        self.price_sns = data.get("price_sns")
        self.variation = data.get("variation")
        self.has_videos = data.get("has_videos")
        self.sales_rank = [
            AmazonProductSalesRank(item) for item in data.get("sales_rank", [])
        ]
        self.top_review = data.get("top_review")
        self.asin_in_url = data.get("asin_in_url")
        self.price_upper = data.get("price_upper")
        self.pricing_str = data.get("pricing_str")
        self.pricing_url = data.get("pricing_url")
        self.discount_end = data.get("discount_end")
        self.manufacturer = data.get("manufacturer")
        self.max_quantity = data.get("max_quantity")
        self.price_buybox = data.get("price_buybox")
        self.product_name = data.get("product_name")
        self.bullet_points = data.get("bullet_points")
        self.is_addon_item = data.get("is_addon_item")
        self.price_initial = data.get("price_initial")
        self.pricing_count = data.get("pricing_count")
        self.reviews_count = data.get("reviews_count")
        self.sns_discounts = data.get("sns_discounts", [])
        self.developer_info = data.get("developer_info", [])
        self.lightning_deal = data.get("lightning_deal")
        self.price_shipping = data.get("price_shipping")
        self.is_prime_pantry = data.get("is_prime_pantry")
        self.product_details = ProductDetails(data.get("product_details", {}))
        self.featured_merchant = data.get("featured_merchant", [])
        self.is_prime_eligible = data.get("is_prime_eligible")
        self.product_dimensions = data.get("product_dimensions")
        self.refurbished_product = AmazonRefurbishedProduct(
            data.get("refurbished_product", {})
        )
        self.answered_questions_count = data.get("answered_questions_count")
        self.rating_star_distribution = [
            AmazonRatingStarDistribution(item)
            for item in data.get("rating_star_distribution", [])
        ]
        self.reviews = [
            AmazonReviews(item) for item in data.get("reviews", [])
        ]
        self.questions = AmazonQuestions(data.get("questions", {}))
        self.questions_total = data.get("questions_total")
        self.business_name = data.get("business_name")
        self.recent_feedback = [
            RecentFeedback(item) for item in data.get("recent_feedback", [])
        ]
        self.business_address = data.get("business_address")
        self.feedback_summary_table = FeedbackSummaryTable(
            data.get("feedback_summary_table", {})
        )
        self.review_count = data.get("review_count")
        self.last_visible_page = data.get("last_visible_page")
        self.parse_status_code = data.get("parse_status_code")


class Result:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.paid = [Paid(item) for item in data.get("paid", [])]
        self.filters = [Filters(item) for item in data.get("filters", [])]
        self.organic = [Organic(item) for item in data.get("organic", [])]
        self.search_information = SearchInformation(
            data.get("search_information")
        )
        self.suggested = [
            SuggestedAmazonSearch(item) for item in data.get("suggested", [])
        ]
        self.amazon_choices = [
            AmazonChoices(item) for item in data.get("amazon_choices", [])
        ]
        self.instant_recommendations = [
            InstantRecommendations(item)
            for item in data.get("instant_recommendations", [])
        ]
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.asin = data.get("asin")
        self.price = data.get("price")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.currency = data.get("currency")
        self.is_prime = data.get("is_prime")
        self.price_str = data.get("price_str")
        self.price_upper = data.get("price_upper")
        self.ratings_count = data.get("ratings_count")


class Paid:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")
        self.data_rw = data.get("data_rw")
        self.data_pcu = data.get("data_pcu")
        self.sitelinks = PaidSitelinks(data.get("sitelinks", {}))
        self.url_shown = data.get("url_shown")
        self.asin = data.get("asin")
        self.price = data.get("price")
        self.rating = data.get("rating")
        self.rel_pos = data.get("rel_pos")
        self.currency = data.get("currency")
        self.url_image = data.get("url_image")
        self.best_seller = data.get("best_seller")
        self.price_upper = data.get("price_upper")
        self.is_sponsored = data.get("is_sponsored")
        self.manufacturer = data.get("manufacturer")
        self.pricing_count = data.get("pricing_count")
        self.reviews_count = data.get("reviews_count")
        self.is_amazons_choice = data.get("is_amazons_choice")
        self.no_price_reason = data.get("no_price_reason")
        self.sales_volume = data.get("sales_volume")
        self.is_prime = data.get("is_prime")
        self.shipping_information = data.get("shipping_information")
        self.pos_overall = data.get("pos_overall")


class PaidSitelinks:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.expanded = [Expanded(item) for item in data.get("expanded", [])]
        self.inline = [Inline(item) for item in data.get("inline", [])]


class Expanded:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")


class Inline:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.desc = data.get("desc")
        self.title = data.get("title")


class Filters:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.name = data.get("name")
        self.values = [FilterValues(item) for item in data.get("values", [])]


class FilterValues:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.value = data.get("value")


class Organic:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.pos = data.get("pos")
        self.url = data.get("url")
        self.type = data.get("type")
        self.price = data.get("price")
        self.title = data.get("title")
        self.currency = data.get("currency")
        self.merchant = Merchant(data.get("merchant", {}))
        self.price_str = data.get("price_str")
        self.product_id = data.get("product_id")
        self.asin = data.get("asin")
        self.rating = data.get("rating")
        self.url_image = data.get("url_image")
        self.best_seller = data.get("best_seller")
        self.price_upper = data.get("price_upper")
        self.is_sponsored = data.get("is_sponsored")
        self.manufacturer = data.get("manufacturer")
        self.pricing_count = data.get("pricing_count")
        self.reviews_count = data.get("reviews_count")
        self.is_amazons_choice = data.get("is_amazons_choice")
        self.no_price_reason = data.get("no_price_reason")
        self.is_prime = data.get("is_prime")
        self.sales_volume = data.get("sales_volume")
        self.variations = [
            Variations(item) for item in data.get("variations", [])
        ]
        self.pos_overall = data.get("pos_overall")


class Merchant:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.name = data.get("name")


class Variations:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.asin = data.get("asin")
        self.title = data.get("title")
        self.price = data.get("price")
        self.price_strikethrough = data.get("price_strikethrough")
        self.not_available = data.get("not_available")


class SearchInformation:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.query = data.get("query")
        self.showing_results_for = data.get("showing_results_for")


class Variants:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.type = data.get("type")
        self.items = [VariantItem(item) for item in data.get("items", [])]


class VariantItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.value = data.get("value")
        self.selected = data.get("selected")
        self.available = data.get("available")
        self.product_id = data.get("product_id")


class RelatedItems:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.items = [RelatedItem(item) for item in data.get("items", [])]


class RelatedItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.price = data.get("price")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.currency = data.get("currency")
        self.reviews_count = data.get("reviews_count")


class Specifications:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.items = [
            SpecificationItem(item) for item in data.get("items", [])
        ]
        self.section_title = data.get("section_title")


class SpecificationItem:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.title = data.get("title")
        self.value = data.get("value")


class Pricing:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.price = data.get("price")
        self.seller = data.get("seller")
        self.details = data.get("details")
        self.currency = data.get("currency")
        self.condition = data.get("condition")
        self.price_tax = data.get("price_tax")
        self.price_total = data.get("price_total")
        self.seller_link = data.get("seller_link")
        self.price_shipping = data.get("price_shipping")
        self.delivery = data.get("delivery")
        self.seller_id = data.get("seller_id")
        self.rating_count = data.get("rating_count")
        self.delivery_options = data.get("delivery_options")


class SuggestedAmazonSearch:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.asin = data.get("asin")
        self.price = data.get("price")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.currency = data.get("currency")
        self.url_image = data.get("url_image")
        self.best_seller = data.get("best_seller")
        self.price_upper = data.get("price_upper")
        self.is_sponsored = data.get("is_sponsored")
        self.manufacturer = data.get("manufacturer")
        self.pricing_count = data.get("pricing_count")
        self.reviews_count = data.get("reviews_count")
        self.is_amazons_choice = data.get("is_amazons_choice")
        self.pos = data.get("pos")
        self.shipping_information = data.get("shipping_information")
        self.sales_volume = data.get("sales_volume")
        self.no_price_reason = data.get("no_price_reason")
        self.suggested_query = data.get("suggested_query")


class AmazonChoices:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.asin = data.get("asin")
        self.price = data.get("price")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.currency = data.get("currency")
        self.url_image = data.get("url_image")
        self.best_seller = data.get("best_seller")
        self.price_upper = data.get("price_upper")
        self.is_sponsored = data.get("is_sponsored")
        self.manufacturer = data.get("manufacturer")
        self.pricing_count = data.get("pricing_count")
        self.reviews_count = data.get("reviews_count")
        self.is_amazons_choice = data.get("is_amazons_choice")
        self.pos = data.get("pos")
        self.is_prime = data.get("is_prime")
        self.shipping_information = data.get("shipping_information")
        self.sales_volume = data.get("sales_volume")
        self.no_price_reason = data.get("no_price_reason")
        self.variations = [
            Variations(item) for item in data.get("variations", [])
        ]


class InstantRecommendations:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.asin = data.get("asin")
        self.price = data.get("price")
        self.title = data.get("title")
        self.rating = data.get("rating")
        self.currency = data.get("currency")
        self.url_image = data.get("url_image")
        self.best_seller = data.get("best_seller")
        self.price_upper = data.get("price_upper")
        self.is_sponsored = data.get("is_sponsored")
        self.manufacturer = data.get("manufacturer")
        self.pricing_count = data.get("pricing_count")
        self.reviews_count = data.get("reviews_count")
        self.is_amazons_choice = data.get("is_amazons_choice")
        self.pos = data.get("pos")
        self.sales_volume = data.get("sales_volume")
        self.no_price_reason = data.get("no_price_reason")


class AmazonProductAds:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.pos = data.get("pos")
        self.asin = data.get("asin")
        self.type = data.get("type")
        self.price = data.get("price")
        self.title = data.get("title")
        self.images = data.get("images", [])
        self.rating = data.get("rating")
        self.location = data.get("location")
        self.price_upper = data.get("price_upper")
        self.reviews_count = data.get("reviews_count")
        self.is_prime_eligible = data.get("is_prime_eligible")


class AmazonProductCategory:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.ladder = [
            {"url": item.get("url"), "name": item.get("name")}
            for item in data.get("ladder", [])
        ]


class AmazonProductDelivery:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.date = Date(data.get("date", {}))
        self.type = data.get("type")


class Date:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.by = data.get("by")
        self.from_date = data.get("from")


class AmazonProductSalesRank:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.rank = data.get("rank")
        self.ladder = [
            {"url": item.get("url"), "name": item.get("name")}
            for item in data.get("ladder", [])
        ]


class ProductDetails:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.asin = data.get("asin")
        self.batteries = data.get("batteries")
        self.item_weight = data.get("item_weight")
        self.manufacturer = data.get("manufacturer")
        self.customer_reviews = data.get("customer_reviews")
        self.best_sellers_rank = data.get("best_sellers_rank")
        self.country_of_origin = data.get("country_of_origin")
        self.item_model_number = data.get("item_model_number")
        self.product_dimensions = data.get("product_dimensions")
        self.date_first_available = data.get("date_first_available")
        self.is_discontinued_by_manufacturer = data.get(
            "is_discontinued_by_manufacturer"
        )


class AmazonRefurbishedProduct:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.link = Link(data.get("link", {}))
        self.condition_title = data.get("condition_title")


class Link:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.url = data.get("url")
        self.title = data.get("title")


class AmazonRatingStarDistribution:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.rating = data.get("rating")
        self.percentage = data.get("percentage")


class AmazonReviews:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.id = data.get("id")
        self.title = data.get("title")
        self.author = data.get("author")
        self.rating = data.get("rating")
        self.content = data.get("content")
        self.timestamp = data.get("timestamp")
        self.is_verified = data.get("is_verified")
        self.product_attributes = data.get("product_attributes")


class AmazonQuestions:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.title = data.get("title")
        self.votes = data.get("votes")
        self.answers = [Answer(item) for item in data.get("answers", [])]


class Answer:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.author = data.get("author")
        self.content = data.get("content")
        self.timestamp = data.get("timestamp")


class RecentFeedback:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.feedback = data.get("feedback")
        self.rated_by = data.get("rated_by")
        self.rating_stars = data.get("rating_stars")


class FeedbackSummaryTable:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.counts = Counts(data.get("counts", {}))
        self.neutral = Counts(data.get("neutral", {}))
        self.negative = Counts(data.get("negative", {}))
        self.positive = Counts(data.get("positive", {}))


class Counts:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.thirty_days = data.get("30_days")
        self.ninety_days = data.get("90_days")
        self.all_time = data.get("all_time")
        self.twelve_months = data.get("12_months")


class Job:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.callback_url = data.get("callback_url")
        self.client_id = data.get("client_id")
        self.context = [Context(item) for item in data.get("context", [])]
        self.created_at = data.get("created_at")
        self.domain = data.get("domain")
        self.geo_location = data.get("geo_location")
        self.id = data.get("id")
        self.limit = data.get("limit")
        self.locale = data.get("locale")
        self.pages = data.get("pages")
        self.parse = data.get("parse")
        self.parser_type = data.get("parser_type")
        self.parsing_instructions = data.get("parsing_instructions")
        self.browser_instructions = data.get("browser_instructions")
        self.render = data.get("render")
        self.url = data.get("url")
        self.query = data.get("query")
        self.source = data.get("source")
        self.start_page = data.get("start_page")
        self.status = data.get("status")
        self.storage_type = data.get("storage_type")
        self.storage_url = data.get("storage_url")
        self.subdomain = data.get("subdomain")
        self.content_encoding = data.get("content_encoding")
        self.updated_at = data.get("updated_at")
        self.user_agent_type = data.get("user_agent_type")
        self.session_info = data.get("session_info")
        self.statuses = data.get("statuses")
        self.client_notes = data.get("client_notes")
        self.links = [JobLink(item) for item in data.get("_links", [])]


class Context:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.key = data.get("key")
        self.value = data.get("value")


class JobLink:
    def __init__(self, data):
        if data is None:
            data = {}
        self.raw = data
        self.rel = data.get("rel")
        self.href = data.get("href")
        self.method = data.get("method")



from typing import Any, List
from urllib.parse import urlparse

import aiohttp

from .defaults import (
    DEFAULT_JOB_COMPLETION_TIMEOUT,
    DEFAULT_POLL_INTERVAL,
    DEFAULT_REQUEST_TIMEOUT,
    DEFAULT_REQUEST_TIMEOUT_ASYNC,
)
from .types import fn_name


def get_valid_values(module: object) -> list:
    """
    Returns a list of valid values from the given module.

    Args:
        module (object): The module to retrieve valid values from.

    Returns:
        list: A list of valid values from the module.
    """
    return [
        getattr(module, name)
        for name in dir(module)
        if not name.startswith("__")
    ]


VALID_FN_NAMES = get_valid_values(fn_name)


def prepare_config(**kwargs):
    """
    Prepare a configuration dictionary based on the provided keyword arguments.

    Args:
        request_timeout (int, optional): The timeout value in seconds. Defaults
        to None.
        poll_interval (int, optional): The poll interval value in seconds.
        Defaults to None.
        job_completion_timeout (int, optional): The job completion timeout
        value in seconds. Defaults to None.

    Returns:
        dict: The prepared configuration dictionary.

    """
    config = {}
    config["request_timeout"] = (
        kwargs["request_timeout"]
        if kwargs.get("request_timeout") is not None
        else (
            DEFAULT_REQUEST_TIMEOUT_ASYNC
            if kwargs.get("async_integration") is not None
            else DEFAULT_REQUEST_TIMEOUT
        )
    )
    config["poll_interval"] = (
        kwargs["poll_interval"]
        if kwargs.get("poll_interval") is not None
        else DEFAULT_POLL_INTERVAL
    )
    config["job_completion_timeout"] = (
        kwargs["job_completion_timeout"]
        if kwargs.get("job_completion_timeout") is not None
        else DEFAULT_JOB_COMPLETION_TIMEOUT
    )

    return config


def validate_url(input_url: str, host: str) -> None:
    """
    Validates if the given URL is valid and belongs to the specified host.

    Args:
        input_url (str): The URL to be validated.
        host (str): The expected domain or host.

    Raises:
        ValueError: If the URL parameter is empty, missing scheme, missing
        host, or does not belong to the specified host.

    Returns:
        None
    """
    # Check if the URL is empty
    if not input_url:
        raise ValueError("URL parameter is empty")

    # Parse the URL
    parsed_url = urlparse(input_url)

    # Check if the scheme (protocol) is present and not empty
    if not parsed_url.scheme:
        raise ValueError("URL is missing scheme")

    # Check if the host is present and not empty
    if not parsed_url.netloc:
        raise ValueError("URL is missing a host")

    # Check if the host matches the expected domain or host
    if host not in parsed_url.netloc:
        raise ValueError(f"URL does not belong to {host}")

    return None


async def ensure_session(session) -> aiohttp.ClientSession:
    """
    Ensure the provided session is valid and return a valid session.

    Args:
        session: The session to ensure.

    Returns:
        A valid aiohttp.ClientSession object.

    """
    if session is None or session.closed:
        session = aiohttp.ClientSession()
    return session


async def close(user_session: aiohttp.ClientSession) -> None:
    """
    Closes the user session.

    Args:
        user_session: The user session to be closed.

    Returns:
        None
    """

    if user_session:
        await user_session.close()


def check_parsing_instructions_validity(instructions: dict) -> None:
    """
    Check the validity of parsing instructions.

    Args:
        instructions (dict): The parsing instructions to be validated.

    Raises:
        Exception: If the parsing instructions have an invalid structure.

    Returns:
        None
    """
    if instructions is None:
        return

    if "_fns" in instructions:
        validate_fns(instructions["_fns"])
    else:
        for key, value in instructions.items():
            if isinstance(value, dict):
                check_parsing_instructions_validity(value)
            else:
                raise Exception(f"Invalid structure for key: {key}")


def validate_fns(fns: List[Any]) -> None:
    """
    Validates a list of functions.

    Args:
        fns (list): A list of functions to validate.

    Raises:
        Exception: If `fns` is None or not a list.
    """
    if fns is None:
        raise Exception("_fns cannot be nil")
    if not isinstance(fns, list):
        raise Exception("_fns must be a list")

    for fn in fns:
        validate_fn(fn)


def validate_fn(fn: dict) -> None:
    """
    Validates the given function dictionary.

    Args:
        fn: A dictionary representing the function.

    Raises:
        ValueError: If the function dictionary is not valid.
    """
    if not isinstance(fn, dict):
        raise ValueError("Each item in _fns must be a dictionary")
    if "_fn" not in fn:
        raise ValueError("_fn must be set in each function")
    if fn["_fn"] not in VALID_FN_NAMES:
        raise ValueError(f"_fn must be a valid function name, got {fn['_fn']}")

    # Delegate to specific argument validators
    validate_fn_args(fn["_fn"], fn.get("_args"))


def validate_fn_args(function: str, args: Any) -> None:
    """
    Validate the arguments for a given function.

    Args:
        function (str): The name of the function to validate.
        args (Any): The arguments to validate.

    Raises:
        ValueError: If there is no validator for the given function name.
    """
    # Map function name to validator function
    validators = {
        fn_name.ELEMENT_TEXT: validate_empty,
        fn_name.LENGTH: validate_empty,
        fn_name.CONVERT_TO_FLOAT: validate_empty,
        fn_name.CONVERT_TO_INT: validate_empty,
        fn_name.CONVERT_TO_STR: validate_empty,
        fn_name.MAX: validate_empty,
        fn_name.MIN: validate_empty,
        fn_name.PRODUCT: validate_empty,
        fn_name.XPATH: validate_string_array,
        fn_name.XPATH_ONE: validate_string_array,
        fn_name.CSS: validate_string_array,
        fn_name.CSS_ONE: validate_string_array,
        fn_name.AMOUNT_FROM_STRING: validate_string,
        fn_name.AMOUNT_RANGE_FROM_STRING: validate_string,
        fn_name.REGEX_FIND_ALL: validate_string,
        fn_name.JOIN: validate_optional_string,
        fn_name.REGEX_SEARCH: validate_list_string_optional_int,
        fn_name.REGEX_SUBSTRING: validate_list_string_optional_int,
        fn_name.SELECT_NTH: validate_non_zero_int,
        fn_name.AVERAGE: validate_optional_int,
    }

    if function not in validators:
        raise ValueError(f"No validator for function name: {function}")

    # Call the appropriate validator
    validator = validators[function]
    validator(args)


def validate_empty(args: Any) -> None:
    """
    Validates if the given argument is empty.

    Args:
        args: The argument to be validated.

    Raises:
        ValueError: If the argument is not empty.
    """
    if args:
        raise ValueError("_args must be empty")


def validate_string_array(args: List[str]) -> None:
    """
    Validates a list of non-empty strings.

    Args:
        args: A list of strings to be validated.

    Raises:
        ValueError: If `args` is not a list of non-empty strings.
    """
    if not isinstance(args, list) or not all(
        isinstance(elem, str) and elem for elem in args
    ):
        raise ValueError("_args must be a list of non-empty strings")


def validate_string(args: str) -> None:
    """
    Validates if the given argument is a non-empty string.

    Args:
        args: The argument to be validated.

    Raises:
        ValueError: If the argument is not a non-empty string.
    """
    if not isinstance(args, str) or not args:
        raise ValueError("_args must be a non-empty string")


def validate_optional_string(args: str) -> None:
    """
    Validates if the given argument is a non-empty string or None.

    Args:
        args (str): The argument to be validated.

    Raises:
        ValueError: If the argument is not a non-empty string or None.
    """
    if args is not None and (not isinstance(args, str) or not args):
        raise ValueError("_args must be a non-empty string or None")


def validate_non_zero_int(args: int) -> None:
    """
    Validates if the given argument is a non-zero integer.

    Args:
        args (int): The argument to be validated.

    Raises:
        ValueError: If the argument is not a non-zero integer.

    Returns:
        None
    """
    if not isinstance(args, int) or args == 0:
        raise ValueError("_args must be a non-zero integer")


def validate_optional_int(args: int) -> None:
    """
    Validates if the given argument is a non-zero integer or None.

    Args:
        args (int): The argument to be validated.

    Raises:
        ValueError: If the argument is not a non-zero integer or None.
    """
    if args is not None and (not isinstance(args, int) or args == 0):
        raise ValueError("_args must be a non-zero integer or None")


def validate_list_string_optional_int(args: list) -> None:
    """
    Validates the input arguments.

    Args:
        args: A list containing the arguments to be validated.

    Raises:
        ValueError: If the first argument is not a non-empty string or if the
        second argument is not a non-zero integer when present.
    """
    if (
        not isinstance(args, list)
        or len(args) < 1
        or not isinstance(args[0], str)
        or not args[0]
    ):
        raise ValueError("_args first argument must be a non-empty string")
    if len(args) > 1 and (not isinstance(args[1], int) or args[1] == 0):
        raise ValueError(
            "_args second argument must be a non-zero integer when present"
        )



import logging

import oxylabs.utils.utils as utils

from .bing.bing import Bing, BingAsync
from .google.google import Google, GoogleAsync
from .response import SERPResponse

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SERP:

    def __init__(self, client) -> None:
        self._client = client
        self.bing = Bing(self)
        self.google = Google(self)

    def _get_resp(self, payload: dict, config: dict) -> dict:
        """
        Processes the payload synchronously and fetches API response.

        Args:
            payload (dict): The payload for the request.
            config (dict): The configuration for the request.

        Returns:
            dict: The response from the server after the job is completed.
        """
        # Remove empty or null values from the payload
        payload = {k: v for k, v in payload.items() if v is not None}

        return SERPResponse(self._client._req(payload, "POST", config))


class SERPAsync:

    def __init__(self, client) -> None:
        self._client = client
        self.bing = BingAsync(self)
        self.google = GoogleAsync(self)
        self._session = None
        self._requests = 0

    async def _get_resp(self, payload: dict, config: dict) -> dict:
        """
        Processes the payload asynchronously and fetches API response.

        Args:
            payload (dict): The payload for the request.
            config (dict): The configuration for the request.

        Returns:
            dict: The response from the server after the job is completed.
        """
        # Remove empty or null values from the payload
        payload = {k: v for k, v in payload.items() if v is not None}

        result: dict = None
        self._requests += 1

        try:
            self._session = await utils.ensure_session(self._session)

            result = await self._client._execute_with_timeout(
                payload, config, self._session
            )
            return SERPResponse(result)

        except Exception as e:
            logger.error(f"An error occurred: {e}")

        finally:
            self._requests -= 1
            if self._requests == 0:
                await utils.close(self._session)
        return SERPResponse(None)



import logging

import oxylabs.utils.utils as utils

from .amazon.amazon import Amazon, AmazonAsync
from .google_shopping.google_shopping import (
    GoogleShopping,
    GoogleShoppingAsync,
)
from .response import EcommerceResponse
from .universal.universal import Universal, UniversalAsync
from .wayfair.wayfair import Wayfair, WayfairAsync

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Ecommerce:

    def __init__(self, client) -> None:
        self._client = client
        self.amazon = Amazon(self)
        self.google_shopping = GoogleShopping(self)
        self.universal = Universal(self)
        self.wayfair = Wayfair(self)

    def _get_resp(self, payload: dict, config: dict) -> dict:
        """
        Processes the payload synchronously and fetches API response.

        Args:
            payload (dict): The payload for the request.
            config (dict): The configuration for the request.

        Returns:
            dict: The response from the server after the job is completed.
        """
        # Remove empty or null values from the payload
        payload = {k: v for k, v in payload.items() if v is not None}

        return EcommerceResponse(self._client._req(payload, "POST", config))


class EcommerceAsync:

    def __init__(self, client) -> None:
        self._client = client
        self.amazon = AmazonAsync(self)
        self.google_shopping = GoogleShoppingAsync(self)
        self.universal = UniversalAsync(self)
        self.wayfair = WayfairAsync(self)
        self._session = None
        self._requests = 0

    async def _get_resp(self, payload: dict, config: dict) -> dict:
        """
        Processes the payload asynchronously and fetches API response.

        Args:
            payload (dict): The payload for the request.
            config (dict): The configuration for the request.

        Returns:
            dict: The response from the server after the job is completed.
        """
        # Remove empty or null values from the payload
        payload = {k: v for k, v in payload.items() if v is not None}

        result = None
        self._requests += 1

        try:

            self._session = await utils.ensure_session(self._session)

            result = await self._client._execute_with_timeout(
                payload, config, self._session
            )
            return EcommerceResponse(result)

        except Exception as e:
            logger.error(f"An error occurred: {e}")

        finally:
            self._requests -= 1
            if self._requests == 0:
                await utils.close(self._session)
        return EcommerceResponse(None)



from typing import Optional

from oxylabs.sources.ecommerce.response import EcommerceResponse
from oxylabs.utils.types import source
from oxylabs.utils.utils import (
    check_parsing_instructions_validity,
    prepare_config,
)


class Universal:
    def __init__(self, ecommerce_instance) -> None:
        """
        Initializes an instance of the Universal class.

        Args:
            ecommerce_instance: The Ecommerce instance associated with the Universal class.
        """
        self._ecommerce_instance = ecommerce_instance

    def scrape_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        geo_location: Optional[str] = None,
        locale: Optional[str] = None,
        render: Optional[str] = None,
        content_encoding: Optional[str] = None,
        context: Optional[list] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parser_type: Optional[str] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: int = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Scrapes Universal search results for a given URL.

        Args:
            url (str): The URL to be scraped.
            user_agent_type (Optional[str]): Device type and browser.
            "geo_location": None,
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            render (Optional[str]): Enables JavaScript rendering.
            content_encoding:  Add this parameter if you are downloading images.
            context: Optional[list],
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            parser_type: Set the value to ecommerce_product to access our AI-powered Adaptive Parser.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(request_timeout=request_timeout)
        payload = {
            "source": source.UNIVERSAL,
            "url": url,
            "user_agent_type": user_agent_type,
            "geo_location": geo_location,
            "locale": locale,
            "render": render,
            "content_encoding": content_encoding,
            "context": context,
            "callback_url": callback_url,
            "parse": parse,
            "parser_type": parser_type,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = self._ecommerce_instance._get_resp(payload, config)
        return response


class UniversalAsync:
    def __init__(self, ecommerce_async_instance) -> None:
        """
        Initializes an instance of the UniversalAsync class.

        Args:
            ecommerce_async_instance: The EcommerceAsync instance associated with the UniversalAsync class.
        """
        self._ecommerce_async_instance = ecommerce_async_instance

    async def scrape_url(
        self,
        url: str,
        user_agent_type: Optional[str] = None,
        geo_location: Optional[str] = None,
        locale: Optional[str] = None,
        render: Optional[str] = None,
        content_encoding: Optional[str] = None,
        context: Optional[list] = None,
        callback_url: Optional[str] = None,
        parse: Optional[bool] = None,
        parser_type: Optional[str] = None,
        parsing_instructions: Optional[dict] = None,
        request_timeout: int = None,
        job_completion_timeout: int = None,
        poll_interval: int = None,
        **kwargs
    ) -> EcommerceResponse:
        """
        Asynchronously scrapes Universal search results for a given URL.

        Args:
            url (str): The URL to be scraped.
            user_agent_type (Optional[str]): Device type and browser.
            "geo_location": None,
            locale (Optional[str]): Accept-Language header value which changes your Bing search
                            page web interface language.
            render (Optional[str]): Enables JavaScript rendering.
            content_encoding:  Add this parameter if you are downloading images.
            context: Optional[list],
            callback_url (Optional[str]): URL to your callback endpoint.
            parse (Optional[bool]): true will return structured data.
            parser_type: Set the value to ecommerce_product to access our AI-powered Adaptive Parser.
            parsing_instructions (Optional[dict]): Instructions for parsing the results.
            request_timeout (int | 165, optional): The interval in seconds for
            the request to time out if no response is returned.
            Defaults to 165.
            poll_interval (int | 5, optional): The interval in seconds to poll
            the server for a response. Defaults to 5
            job_completion_timeout (int | 50, optional): The interval in
            seconds for the job to time out if no response is returned.
            Defaults to 50.

        Returns:
            EcommerceResponse: The response from the server after the job is completed.
        """

        config = prepare_config(
            request_timeout=request_timeout,
            poll_interval=poll_interval,
            job_completion_timeout=job_completion_timeout,
            async_integration=True,
        )
        payload = {
            "source": source.UNIVERSAL,
            "url": url,
            "user_agent_type": user_agent_type,
            "geo_location": geo_location,
            "locale": locale,
            "render": render,
            "content_encoding": content_encoding,
            "context": context,
            "callback_url": callback_url,
            "parse": parse,
            "parser_type": parser_type,
            "parsing_instructions": parsing_instructions,
            **kwargs,
        }
        check_parsing_instructions_validity(parsing_instructions)
        response = await self._ecommerce_async_instance._get_resp(
            payload, config
        )
        return response



MOBILE = "mobile"
TABLET = "tablet"
DESKTOP = "desktop"
MOBILE_IOS = "mobile_ios"
TABLET_IOS = "tablet_ios"
DESKTOP_EDGE = "desktop_edge"
DESKTOP_OPERA = "desktop_opera"
DESKTOP_SAFARI = "desktop_safari"
MOBILE_ANDROID = "mobile_android"
DESKTOP_CHROME = "desktop_chrome"
TABLET_ANDROID = "tablet_android"
DESKTOP_FIREFOX = "desktop_firefox"



from setuptools import find_packages, setup


def get_version(rel_path):
    with open(rel_path, "r") as file:
        for line in file.read().splitlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


long_description = """
The official Python library for the Oxylabs Scraper APIs.

Collect public data at scale with industry-leading web scraping solutions and the world's largest ethical proxy network.
Documentation can be found at https://github.com/oxylabs/oxylabs-sdk-python.
"""

setup(
    name="oxylabs",
    version=get_version("src/oxylabs/_version.py"),
    description="Official Python library for Oxylabs Scraper APIs",
    long_description=long_description,
    url="https://oxylabs.io/",
    author="Oxylabs",
    author_email="support@oxylabs.io",
    license="MIT",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=["aiohttp", "requests"],
)



RU = "ru"
UA = "ua"
TR = "tr"
CN = "cn"
COM_AI = "com.ai"
COM_PR = "com.pr"
SR = "sr"
ML = "ml"
COM_LB = "com.lb"
BF = "bf"
FM = "fm"
COM_MX = "com.mx"
BJ = "bj"
EE = "ee"
MV = "mv"
NE = "ne"
AT = "at"
GG = "gg"
AE = "ae"
CO_UZ = "co.uz"
AM = "am"
COM_SA = "com.sa"
TL = "tl"
COM_NA = "com.na"
COM_BH = "com.bh"
DK = "dk"
COM_SB = "com.sb"
RO = "ro"
BY = "by"
COM_CO = "com.co"
COM_GI = "com.gi"
CO_ID = "co.id"
MS = "ms"
COM_NG = "com.ng"
IS = "is"
COM_EG = "com.eg"
COM_ET = "com.et"
COM_AF = "com.af"
CH = "ch"
CO_AO = "co.ao"
CL = "cl"
CO_ZA = "co.za"
COM_NF = "com.nf"
DK_RO = "ro"
MD = "md"
ES = "es"
BJ_YO = "bj"
HU = "hu"
DJ = "dj"
COM_MT = "com.mt"
COM_EC = "com.ec"
CO_IN = "co.in"
LK = "lk"
CO_KE = "co.ke"
GY = "gy"
BE = "be"
VG = "vg"
CO_BW = "co.bw"
COM_VN = "com.vn"
CO_TZ = "co.tz"
NE_HA = "ne"
CO_ZW = "co.zw"
TO = "to"
KZ = "kz"
COM_UY = "com.uy"
IQ = "iq"
COM_TW = "com.tw"
RW = "rw"
AD = "ad"
COM_LY = "com.ly"
AL = "al"
CO_IL = "co.il"
KI = "ki"
COM = "com"
MU = "mu"
SC = "sc"
COM_HK = "com.hk"
COM_PA = "com.pa"
CA = "ca"
GE = "ge"
COM_GT = "com.gt"
LI = "li"
COM_KH = "com.kh"
CO_CR = "co.cr"
COM_BO = "com.bo"
CO_VE = "co.ve"
COM_NI = "com.ni"
TD = "td"
CF = "cf"
TK = "tk"
BI = "bi"
MG = "mg"
COM_BD = "com.bd"
COM_BZ = "com.bz"
GM = "gm"
LA = "la"
COM_KW = "com.kw"
CM = "cm"
HT = "ht"
NO = "no"
COM_FJ = "com.fj"
TM = "tm"
COM_SL = "com.sl"
COM_MM = "com.mm"
IM = "im"
SI = "si"
COM_QA = "com.qa"
COM_PE = "com.pe"
CD = "cd"
TT = "tt"
COM_TR = "com.tr"
TG = "tg"
CO_LS = "co.ls"
GR = "gr"
GL = "gl"
MK = "mk"
CO_ZM = "co.zm"
COM_PH = "com.ph"
IT = "it"
CO_JP = "co.jp"
WS = "ws"
COM_AR = "com.ar"
CO_MZ = "co.mz"
AZ = "az"
CO_CK = "co.ck"
FI = "fi"
COM_BN = "com.bn"
PT = "pt"
COM_TJ = "com.tj"
COM_CY = "com.cy"
CV = "cv"
COM_MY = "com.my"
IE = "ie"
COM_SG = "com.sg"
DE = "de"
BA = "ba"
LU = "lu"
BG = "bg"
CO_VI = "co.vi"
COM_OM = "com.om"
AS = "as"
DZ = "dz"
FR = "fr"
LV = "lv"
LT = "lt"
PS = "ps"
SE = "se"
CG = "cg"
NR = "nr"
CO_UG = "co.ug"
COM_VC = "com.vc"
JO = "jo"
CO_TH = "co.th"
RS = "rs"
BS = "bs"
COM_PK = "com.pk"
CO_UK = "co.uk"
SO = "so"
GA = "ga"
COM_UA = "com.ua"
HR = "hr"
COM_CU = "com.cu"
SK = "sk"
COM_NP = "com.np"
NU = "nu"
MN = "mn"
VU = "vu"
NL = "nl"
PT_ST = "st"
COM_BR = "com.br"
TH = "co.th"
MW = "mw"
COM_PG = "com.pg"
PL = "pl"
CO_NZ = "co.nz"
KG = "kg"
CI = "ci"
SH = "sh"
COM_DO = "com.do"
SN = "sn"
COM_JM = "com.jm"
CO_MA = "co.ma"
COM_TN = "com.tn"
DM = "dm"
COM_SV = "com.sv"
COM_SG_2 = "com.sg"
GP = "gp"
ME = "me"
COM_AG = "com.ag"
CZ = "cz"
COM_PY = "com.py"
MR_IN = "co.in"
COM_GH = "com.gh"
ST_LS = "co.ls"
BT = "bt"
RU_KZ = "kz"
IT_SM = "sm"
JE = "je"
TN = "tn"
COM_AU = "com.au"
ME_ME = "me"
PN = "pn"
HN = "hn"
CO_KR = "co.kr"
AR = "com.ar"
BO = "com.bo"
BZ = "com.bz"
UY = "com.uy"
COM_VE = "com.ve"
ID_TL = "tl"



ELEMENT_TEXT = "element_text"
XPATH = "xpath"
XPATH_ONE = "xpath_one"
CSS = "css"
CSS_ONE = "css_one"
AMOUNT_FROM_STRING = "amount_from_string"
AMOUNT_RANGE_FROM_STRING = "amount_range_from_string"
JOIN = "join"
REGEX_FIND_ALL = "regex_find_all"
REGEX_SEARCH = "regex_search"
REGEX_SUBSTRING = "regex_substring"
LENGTH = "length"
SELECT_NTH = "select_nth"
CONVERT_TO_FLOAT = "convert_to_float"
CONVERT_TO_INT = "convert_to_int"
CONVERT_TO_STR = "convert_to_str"
AVERAGE = "average"
MAX = "max"
MIN = "min"
PRODUCT = "product"
